//! # Changelog v0.6
//!
//! * Add more distributions for random sampling.
//! * Fix float aggregations with NaNs.
//! * Comparisons are more performant.
//! * Outer join is more performant.
//! * Start with parallel iterator support for ChunkedArrays.
//! * Remove crossbeam dependency.
